@@ -1,42 +1,84 @@
 from pathlib import Path
-from traceback import print_exc
-from typing import Literal, Union, Text, Optional, Tuple
+from typing import Union, Text, Optional, Tuple
 
 import matplotlib.pyplot as plt
 from pyannote.core import Annotation, Segment, SlidingWindowFeature, notebook
 from pyannote.database.util import load_rttm
 from pyannote.metrics.diarization import DiarizationErrorRate
 from rx.core import Observer
+from typing_extensions import Literal
+from traceback import print_exc
+
+
+class WindowClosedException(Exception):
+    pass
+
+
+def _extract_annotation(value: Union[Tuple, Annotation]) -> Annotation:
+    if isinstance(value, tuple):
+        return value[0]
+    if isinstance(value, Annotation):
+        return value
+    msg = f"Expected tuple or Annotation, but got {type(value)}"
+    raise ValueError(msg)
 
 
 class RTTMWriter(Observer):
     def __init__(self, path: Union[Path, Text], patch_collar: float = 0.05):
         super().__init__()
         self.patch_collar = patch_collar
-        self.path = Path(path)
+        self.path = Path(path).expanduser()
         if self.path.exists():
             self.path.unlink()
 
-    def patch_rttm(self):
+    def patch(self):
         """Stitch same-speaker turns that are close to each other"""
-        annotation = list(load_rttm(self.path).values())[0]
-        with open(self.path, 'w') as file:
-            annotation.support(self.patch_collar).write_rttm(file)
+        annotations = list(load_rttm(self.path).values())
+        if annotations:
+            with open(self.path, 'w') as file:
+                annotations[0].support(self.patch_collar).write_rttm(file)
 
-    def on_next(self, value: Tuple[Annotation, Optional[SlidingWindowFeature]]):
+    def on_next(self, value: Union[Tuple, Annotation]):
+        annotation = _extract_annotation(value)
         with open(self.path, 'a') as file:
-            value[0].write_rttm(file)
+            annotation.write_rttm(file)
 
     def on_error(self, error: Exception):
-        try:
-            self.patch_rttm()
-        except Exception:
-            print("Error while patching RTTM file:")
-            print_exc()
-            exit(1)
+        self.patch()
+        print_exc()
 
     def on_completed(self):
-        self.patch_rttm()
+        self.patch()
+
+
+class DiarizationPredictionAccumulator(Observer):
+    def __init__(self, patch_collar: float = 0.05):
+        super().__init__()
+        self.patch_collar = patch_collar
+        self._annotation = None
+
+    def patch(self):
+        """Stitch same-speaker turns that are close to each other"""
+        self._annotation.support(self.patch_collar)
+
+    def get_prediction(self) -> Annotation:
+        # Patch again in case this is called before on_completed
+        self.patch()
+        return self._annotation
+
+    def on_next(self, value: Union[Tuple, Annotation]):
+        annotation = _extract_annotation(value)
+        if self._annotation is None:
+            self._annotation = annotation
+        else:
+            self._annotation.update(annotation)
+
+    def on_error(self, error: Exception):
+        self.patch()
+        print_exc()
+
+    def on_completed(self):
+        self.patch()
 
 
 class RealTimePlot(Observer):
@@ -56,20 +98,23 @@ class RealTimePlot(Observer):
         self.window_duration = duration
         self.latency = latency
         self.figure, self.axs, self.num_axs = None, None, -1
+        self.window_closed = False
 
-    def _init_num_axs(self, waveform: Optional[SlidingWindowFeature]):
+    def _on_window_closed(self, event):
+        self.window_closed = True
+
+    def _init_num_axs(self):
         if self.num_axs == -1:
-            self.num_axs = 1
-            if waveform is not None:
-                self.num_axs += 1
+            self.num_axs = 2
             if self.reference is not None:
                 self.num_axs += 1
 
-    def _init_figure(self, waveform: Optional[SlidingWindowFeature]):
-        self._init_num_axs(waveform)
+    def _init_figure(self):
+        self._init_num_axs()
         self.figure, self.axs = plt.subplots(self.num_axs, 1, figsize=(10, 2 * self.num_axs))
         if self.num_axs == 1:
             self.axs = [self.axs]
+        self.figure.canvas.mpl_connect('close_event', self._on_window_closed)
 
     def _clear_axs(self):
         for i in range(self.num_axs):
@@ -83,10 +128,13 @@ class RealTimePlot(Observer):
         return Segment(start_time, end_time)
 
     def on_next(self, values: Tuple[Annotation, SlidingWindowFeature, float]):
+        if self.window_closed:
+            raise WindowClosedException
+
         prediction, waveform, real_time = values
         # Initialize figure if first call
         if self.figure is None:
-            self._init_figure(waveform)
+            self._init_figure()
         # Clear previous plots
         self._clear_axs()
         # Set plot bounds
@@ -99,16 +147,9 @@ class RealTimePlot(Observer):
             prediction.rename_labels(mapping=mapping, copy=False)
         notebook.plot_annotation(prediction, self.axs[0])
         self.axs[0].set_title("Output")
-        if self.num_axs == 2:
-            if waveform is not None:
-                notebook.plot_feature(waveform, self.axs[1])
-                self.axs[1].set_title("Audio")
-            elif self.reference is not None:
-                notebook.plot_annotation(self.reference, self.axs[1])
-                self.axs[1].set_title("Reference")
-        elif self.num_axs == 3:
-            notebook.plot_feature(waveform, self.axs[1])
-            self.axs[1].set_title("Audio")
+        notebook.plot_feature(waveform, self.axs[1])
+        self.axs[1].set_title("Audio")
+        if self.num_axs == 3:
             notebook.plot_annotation(self.reference, self.axs[2])
             self.axs[2].set_title("Reference")
 
@@ -119,5 +160,5 @@ class RealTimePlot(Observer):
         plt.pause(0.05)
 
     def on_error(self, error: Exception):
-        print_exc()
-        exit(1)
+        if not isinstance(error, WindowClosedException):
+            print_exc()
